/* (C) 2009-2010,2012 Yana */

#include "device.h"
#include "y_arm.h"
#include "y_base.h"
#include "y_device.h"
#include "yk0a.h"
#include "yk0a_interrupt.h"
#include <stddef.h>

y_uint32_t glue_arm_read_control(void) {
	y_uint32_t control;
	__asm__ __volatile__("mrc p15, 0, %0, c1, c0, 0" : "=r"(control));
	return control;
}

y_uint32_t glue_arm_read_status(void) {
	y_uint32_t status;
	__asm__ __volatile__("mrc p15, 0, %0, c5, c0, 0" : "=r"(status));
	return status;
}

y_uint32_t glue_arm_read_address(void) {
	y_uint32_t address;
	__asm__ __volatile__("mrc p15, 0, %0, c6, c0, 0" : "=r"(address));
	return address;
}

void glue_arm_write_control(y_uint32_t control) {
	__asm__ __volatile__("mcr p15, 0, %0, c1, c0, 0" : : "r"(control));
}

void glue_arm_write_translation(y_uint32_t translation) {
	__asm__ __volatile__("mcr p15, 0, %0, c2, c0, 0" : : "r"(translation));
}

void glue_arm_write_domain(y_uint32_t domain) {
	__asm__ __volatile__("mcr p15, 0, %0, c3, c0, 0" : : "r"(domain));
}

void glue_arm_write_status(y_uint32_t status) {
	__asm__ __volatile__("mcr p15, 0, %0, c5, c0, 0" : : "r"(status));
}

void glue_arm_tlb_invalidate(void) {
	__asm__ __volatile__("mcr p15, 0, %0, c8, c7, 0" : : "r"(0));
}

#include "hw_intr.h"
#include "kernel.h"
#include "vm.h"
#include <signal.h>
#include <string.h>
#include <machine/vm.h>

#define DEVICE_PHYS 0x10000
#define DEVICE_SIZE 0x10000 /* 256 MB */
#define INTERRUPT_BIT (DEVICE_INTERRUPT_RTC | DEVICE_INTERRUPT_ETHER | DEVICE_INTERRUPT_CONS)
#define INTERRUPT_SIZE 8
#define LEVEL1_SIZE 4096 /* 4096 MB */
#define LEVEL1_UNIT 0x100000
#define LEVEL2_SIZE (2048 * 256) /* 2048 MB */
#define STACK_FRAME 17

#if !defined(MINIX_PROCESS) || MINIX_PROCESS % LEVEL1_UNIT != 0
#error MINIX_PROCESS
#endif

const y_uint8_t glue_boot[] = "memory=00100000:01f00000";

irq_hook_t glue_irq;
volatile yk0a_interrupt_t glue_mask;
struct proc *glue_proc;
y_uint32_t glue_vmctl;

y_uint32_t glue_level1[LEVEL1_SIZE] __attribute__((aligned(16384)));
y_uint32_t glue_level2[LEVEL2_SIZE] __attribute__((aligned(1024)));

void glue_exception(struct proc *p, int signal) {
	if (iskernelp(p)) {
		panic("glue_exception");
	} else {
		cause_sig(proc_nr(p), signal);
	}
}

void glue_interrupt(yk0a_interrupt_t ignition) {
	yk0a_interrupt_t i;
	for (i = 0; i < INTERRUPT_SIZE; ++i) {
		if ((1 << (INTERRUPT_SIZE - 1 - i) & ignition) != 0) {
			irq_handle(INTERRUPT_SIZE - 1 - i);
		}
	}
}

void glue_pagefault(struct proc *p, y_uint32_t address, y_uint32_t status) {
	if ((p->p_endpoint <= INIT_PROC_NR && (p->p_misc_flags & MF_FULLVM) == 0) || p->p_endpoint == VM_PROC_NR) {
		cause_sig(proc_nr(p), SIGSEGV);
	} else {
		message m;
		RTS_SET(p, RTS_PAGEFAULT);
		m.m_source = p->p_endpoint;
		m.m_type = VM_PAGEFAULT;
		m.VPF_ADDR = address;
		if ((status & 0xf) != 0xf) {
			m.VPF_FLAGS = 0;
		} else {
			m.VPF_FLAGS = I386_VM_PFE_W;
		}
		if (mini_send(p, VM_PROC_NR, &m, FROM_KERNEL) != 0) {
			panic("glue_pagefault");
		}
	}
}

void glue_task(phys_bytes a) {
	y_uint32_t i;
	for (i = 0; i < LEVEL1_SIZE; ++i) {
		if (i * (LEVEL1_UNIT / CLICK_SIZE) >= DEVICE_PHYS && i * (LEVEL1_UNIT / CLICK_SIZE) < DEVICE_PHYS + DEVICE_SIZE) {
			glue_level1[i] = i << 20 | 0x422; /* AP = 1, domain = 1, C = 0, type = 2 */
		} else if (i * (LEVEL1_UNIT / CLICK_SIZE) >= MINIX_PROCESS / CLICK_SIZE && i * (LEVEL1_UNIT / CLICK_SIZE) < MINIX_PROCESS / CLICK_SIZE + LEVEL2_SIZE) {
			glue_level1[i] = (y_uint32_t)&glue_level2[i * (LEVEL1_UNIT / CLICK_SIZE) - MINIX_PROCESS / CLICK_SIZE] | 1; /* domain = 0, type = 1 */
		} else {
			glue_level1[i] = i << 20 | 0x40a; /* AP = 1, domain = 0, C = 1, type = 2 */
		}
	}
	glue_arm_write_translation((y_uint32_t)glue_level1);
	glue_arm_write_domain(0x5); /* D1 = 1, D0 = 1 */
	glue_arm_write_control(glue_arm_read_control() | 1); /* M = 1 */
	glue_arm_write_status(0);
	for (i = 0; i < NR_BOOT_PROCS; ++i) {
		if (i < NR_TASKS) {
			image[i].memmap.text_vaddr = ((struct exec *)a)[0].a_syms;
			image[i].memmap.text_paddr = ((struct exec *)a)[0].a_syms;
			image[i].memmap.text_bytes = (((struct exec *)a)[0].a_text + CLICK_SIZE - 1) / CLICK_SIZE * CLICK_SIZE;
			image[i].memmap.data_vaddr = image[i].memmap.text_vaddr + image[i].memmap.text_bytes;
			image[i].memmap.data_paddr = image[i].memmap.text_paddr + image[i].memmap.text_bytes;
			image[i].memmap.data_bytes = (((struct exec *)a)[0].a_data + ((struct exec *)a)[0].a_bss + CLICK_SIZE - 1) / CLICK_SIZE * CLICK_SIZE;
			image[i].memmap.stack_bytes = (((struct exec *)a)[0].a_total + CLICK_SIZE - 1) / CLICK_SIZE * CLICK_SIZE - image[i].memmap.data_bytes;
			image[i].memmap.entry = ((struct exec *)a)[0].a_entry;
		} else {
			image[i].memmap.text_vaddr = MINIX_PROCESS;
			image[i].memmap.text_paddr = ((struct exec *)a)[i - NR_TASKS + 1].a_syms;
			image[i].memmap.text_bytes = (((struct exec *)a)[i - NR_TASKS + 1].a_text + CLICK_SIZE - 1) / CLICK_SIZE * CLICK_SIZE;
			image[i].memmap.data_vaddr = image[i].memmap.text_vaddr + image[i].memmap.text_bytes;
			image[i].memmap.data_paddr = image[i].memmap.text_paddr + image[i].memmap.text_bytes;
			image[i].memmap.data_bytes = (((struct exec *)a)[i - NR_TASKS + 1].a_data + ((struct exec *)a)[i - NR_TASKS + 1].a_bss + CLICK_SIZE - 1) / CLICK_SIZE * CLICK_SIZE;
			image[i].memmap.stack_bytes = (((struct exec *)a)[i - NR_TASKS + 1].a_total + CLICK_SIZE - 1) / CLICK_SIZE * CLICK_SIZE - image[i].memmap.data_bytes;
			image[i].memmap.entry = ((struct exec *)a)[i - NR_TASKS + 1].a_entry;
		}
	}
	cstart(0, 0, 0, 0, 0);
	main();
	yk0a_set_context(2, YK0A_CONTEXT_CPSR, 0, NULL);
	while (1) {
		yk0a_flag_t f;
		yk0a_interrupt_t ignition;
		struct proc *p;
		yk0a_task_t t;
		p = get_cpulocal_var(proc_ptr);
		if (p != glue_proc || glue_vmctl != 0) {
			glue_proc = p;
			glue_vmctl = 0;
			glue_arm_tlb_invalidate();
			/* ToDo: cache */
			if (p->p_seg.p_cr3 == 0) {
				for (i = 0; i < LEVEL2_SIZE; ++i) {
					if (i + MINIX_PROCESS / CLICK_SIZE >= p->p_memmap[T].mem_vir && i + MINIX_PROCESS / CLICK_SIZE < p->p_memmap[T].mem_vir + p->p_memmap[T].mem_len) {
						glue_level2[i] = (i + MINIX_PROCESS / CLICK_SIZE + p->p_memmap[T].mem_phys - p->p_memmap[T].mem_vir) * CLICK_SIZE | 0xaaa; /* AP3 = 2, AP2 = 2, AP1 = 2, AP0 = 2, C = 1, type = 2 */
					} else if (i + MINIX_PROCESS / CLICK_SIZE >= p->p_memmap[D].mem_vir && i + MINIX_PROCESS / CLICK_SIZE < p->p_memmap[S].mem_vir + p->p_memmap[S].mem_len) {
						glue_level2[i] = (i + MINIX_PROCESS / CLICK_SIZE + p->p_memmap[D].mem_phys - p->p_memmap[D].mem_vir) * CLICK_SIZE | 0xffa; /* AP3 = 3, AP2 = 3, AP1 = 3, AP0 = 3, C = 1, type = 2 */
					} else {
						glue_level2[i] = 0; /* type = 0 */
					}
				}
			} else {
				for (i = 0; i < LEVEL2_SIZE; ++i) {
					y_uint32_t d;
					d = *(y_uint32_t *)(p->p_seg.p_cr3 + I386_VM_PDE(i * CLICK_SIZE + MINIX_PROCESS) * 4);
					if ((d & I386_VM_PRESENT) == 0) {
						glue_level2[i] = 0; /* type = 0 */
					} else {
						y_uint32_t t;
						t = *(y_uint32_t *)(I386_VM_PFA(d) + I386_VM_PTE(i * CLICK_SIZE + MINIX_PROCESS) * 4);
						if ((t & I386_VM_PRESENT) == 0) {
							glue_level2[i] = 0; /* type = 0 */
						} else if ((t & I386_VM_WRITE) == 0) {
							glue_level2[i] = I386_VM_PFA(t) | 0xaaa; /* AP3 = 2, AP2 = 2, AP1 = 2, AP0 = 2, C = 1, type = 2 */
						} else {
							glue_level2[i] = I386_VM_PFA(t) | 0xffa; /* AP3 = 3, AP2 = 3, AP1 = 3, AP0 = 3, C = 1, type = 2 */
						}
					}
				}
			}
#if defined(YK0A_CUSTOM_ENABLE_PORT)
			if (iskernelp(p)) {
				yk0a_set_context(2, YK0A_CUSTOM_CONTEXT_PORT, ~(yk0a_port_t)0, NULL);
			} else {
				yk0a_set_context(2, YK0A_CUSTOM_CONTEXT_PORT, 1, NULL);
			}
#endif
		}
		for (i = 0; i < STACK_FRAME; ++i) {
			yk0a_set_context(2, i, ((yk0a_context_value_t *)&p->p_reg)[i], NULL);
		}
		yk0a_reply(2);
		yk0a_listen(1, YK0A_FLAG_KERNEL_SENDER | glue_mask, &f);
		ignition = f & glue_mask;
		if ((f & YK0A_FLAG_KERNEL_SENDER) == 0) {
			yk0a_forward(1, 2);
		}
		if (yk0a_accept(1, &t) != 0 || t != 2) {
y_device_cons_put_string("{accept}\n");
while (1) {
}
		}
		for (i = 0; i < STACK_FRAME; ++i) {
			yk0a_get_context(2, i, &((yk0a_context_value_t *)&p->p_reg)[i]);
		}
		if ((f & YK0A_FLAG_KERNEL_SENDER) != 0) {
			y_uint32_t address;
			yk0a_context_value_t cpsr;
			y_uint32_t status;
			yk0a_set_context(2, YK0A_CONTEXT_CPSR, 0, &cpsr);
			switch ((cpsr & Y_ARM_CPSR_M_MASK) >> Y_ARM_CPSR_M_SHIFT) {
			case 0:
				context_stop(p);
#if defined(YK0A_CUSTOM_ENABLE_PORT)
				if (iskernelp(p)) {
					if (p->p_reg.r2 == 0) {
						kernel_call((message *)p->p_reg.r3, p);
					} else {
						p->p_reg.retreg = do_ipc(p->p_reg.r2, p->p_reg.r3, p->p_reg.r4);
					}
				} else {
					if (p->p_reg.retreg == 0) {
						kernel_call((message *)p->p_reg.r1, p);
					} else {
						p->p_reg.retreg = do_ipc(p->p_reg.retreg, p->p_reg.r1, p->p_reg.r2);
					}
				}
#else
				if (p->p_reg.r2 == 0) {
					kernel_call((message *)p->p_reg.r3, p);
				} else {
					p->p_reg.retreg = do_ipc(p->p_reg.r2, p->p_reg.r3, p->p_reg.r4);
				}
#endif
				switch_to_user();
				break;
			case 0x17: /* abt */
				status = glue_arm_read_status();
				if (status == 0) {
					p->p_reg.pc -= 4;
					address = p->p_reg.pc;
				} else {
					p->p_reg.pc -= 8;
					address = glue_arm_read_address();
					glue_arm_write_status(0);
				}
				context_stop(p);
				glue_pagefault(p, address, status);
				switch_to_user();
				break;
			case 0x1b: /* und */
				context_stop(p);
				glue_exception(p, SIGILL);
				switch_to_user();
				break;
			default:
y_device_cons_put_string("{exception}\n");
while (1) {
}
				break;
			}
		}
		if (ignition != 0) {
			context_stop(p);
			glue_interrupt(ignition);
			switch_to_user();
		}
	}
}

#include "arch_proto.h"

__dead void reset(void) {
y_device_cons_put_string("{reset}\n");
while (1) {
}
}

void phys_memset(phys_bytes ph, u32_t c, phys_bytes bytes) {
	/* ToDo: cache */
	memset((void *)ph, c, bytes);
}

void __switch_address_space(struct proc * p, struct proc ** __ptproc) {
	/* ToDo: MMU */
}

#include "clock.h"

int init_local_timer(unsigned freq) {
	*(volatile y_uint32_t *)DEVICE_RTC_HZ = freq;
	return 0;
}

void stop_local_timer(void) {
y_device_cons_put_string("{stop_local_timer}\n");
while (1) {
}
}

void restart_local_timer(void) {
}

int register_local_timer_handler(irq_handler_t handler) {
	glue_irq.proc_nr_e = NONE;
	put_irq_handler(&glue_irq, CLOCK_IRQ, handler);
	return 0;
}

u64_t ms_2_cpu_time(unsigned ms) {
	/* ToDo: TSC */
	return ms;
}

unsigned cpu_time_2_ms(u64_t cpu_time) {
y_device_cons_put_string("{cpu_time_2_ms}\n");
while (1) {
}
}

#include "hw_intr.h"

void irq_8259_unmask(int irq) {
#if 0
	if (1 << irq == DEVICE_INTERRUPT_RTC) {
		return;
	}
#else /* patch */
#endif
	yk0a_interrupt_bit_mask(INTERRUPT_BIT, 0, NULL);
	glue_mask |= 1 << irq & INTERRUPT_BIT;
	yk0a_interrupt_bit_mask(0, glue_mask, NULL);
}

void irq_8259_mask(int irq) {
	yk0a_interrupt_bit_mask(INTERRUPT_BIT, 0, NULL);
	glue_mask &= ~(1 << irq & INTERRUPT_BIT);
	yk0a_interrupt_bit_mask(0, glue_mask, NULL);
}

void irq_8259_eoi(int irq) {
	if (1 << irq == DEVICE_INTERRUPT_RTC) {
		*(volatile y_uint32_t *)DEVICE_RTC_INTERRUPT_ACK;
	}
}

#include "proto.h"

void cycles_accounting_init(void) {
	/* ToDo: TSC */
}

void context_stop(struct proc * p) {
	/* ToDo: TSC */
}

void restore_fpu(struct proc *pr) {
y_device_cons_put_string("{restore_fpu}\n");
while (1) {
}
}

void save_fpu(struct proc *pr) {
	/* ToDo: FPU */
}

void save_local_fpu(struct proc *pr) {
y_device_cons_put_string("{save_local_fpu}\n");
while (1) {
}
}

void fpu_sigcontext(struct proc *pr, struct sigframe *fr, struct sigcontext *sc) {
	/* ToDo: FPU */
}

struct proc * arch_finish_switch_to_user(void) {
	return get_cpulocal_var(proc_ptr);
}

void prot_init(void) {
}

#if 0
phys_bytes phys_copy(phys_bytes source, phys_bytes dest, phys_bytes count) {
y_device_cons_put_string("{phys_copy}\n");
while (1) {
}
}
#else /* patch */
#endif

#if 0
void phys_copy_fault(void) {
y_device_cons_put_string("{phys_copy_fault}\n");
while (1) {
}
}
#else /* patch */
#endif

#if 0
void phys_copy_fault_in_kernel(void) {
y_device_cons_put_string("{phys_copy_fault_in_kernel}\n");
while (1) {
}
}
#else /* patch */
#endif

int virtual_copy_f(struct proc * caller, struct vir_addr *src, struct vir_addr *dst, vir_bytes bytes, int vmcheck) {
	/* ToDo: cache */
	y_uint32_t i;
	phys_bytes l[2];
	struct proc *p[2];
	struct vir_addr *v[2];
	v[0] = src;
	v[1] = dst;
	for (i = 0; i < 2; ++i) {
		proc_nr_t n;
		switch (v[i]->segment & SEGMENT_TYPE) {
		case LOCAL_SEG:
			if (!isokendpt(v[i]->proc_nr_e, &n)) {
				return EDEADSRCDST;
			}
			p[i] = proc_addr(n);
			l[i] = umap_local(p[i], v[i]->segment & SEGMENT_INDEX, v[i]->offset, bytes);
			if (l[i] == 0) {
				return EFAULT;
			}
			break;
		case PHYS_SEG:
			p[i] = NULL;
			l[i] = v[i]->offset;
			break;
		default:
y_device_cons_put_string("{virtual_copy_f}\n");
while (1) {
}
			break;
		}
	}
	if (!vm_running) {
		memcpy((void *)l[1], (void *)l[0], bytes);
		return 0;
	} else {
		phys_bytes o;
		phys_bytes s[2];
		if (caller != NULL && (caller->p_misc_flags & MF_KCALL_RESUME) != 0 && caller->p_vmrequest.vmresult != 0) {
			return caller->p_vmrequest.vmresult;
		}
		s[0] = 0;
		s[1] = 0;
		o = 0;
		while (o < bytes) {
			phys_bytes n;
			for (i = 0; i < 2; ++i) {
				if (s[i] == 0) {
					if (p[i] == NULL || p[i]->p_seg.p_cr3 == 0) {
						s[i] = bytes;
					} else {
						y_uint32_t d;
						y_uint32_t t;
						d = *(y_uint32_t *)(p[i]->p_seg.p_cr3 + I386_VM_PDE(v[i]->offset + o) * 4);
						if ((d & I386_VM_PRESENT) == 0) {
							break;
						}
						t = *(y_uint32_t *)(I386_VM_PFA(d) + I386_VM_PTE(v[i]->offset + o) * 4);
						if ((t & I386_VM_PRESENT) == 0) {
							break;
						}
						if ((t & I386_VM_WRITE) == 0 && i == 1) {
							break;
						}
						l[i] = I386_VM_PFA(t) + (v[i]->offset + o) % I386_PAGE_SIZE;
						s[i] = I386_PAGE_SIZE - (v[i]->offset + o) % I386_PAGE_SIZE;
						if (s[i] > bytes - o) {
							s[i] = bytes - o;
						}
					}
				}
			}
			if (i != 2) {
				if (caller == NULL || vmcheck == 0) {
					if (i == 0) {
						return EFAULT_SRC;
					} else {
						return EFAULT_DST;
					}
				}
				RTS_SET(caller, RTS_VMREQUEST);
				caller->p_vmrequest.type = VMSTYPE_KERNELCALL;
				caller->p_vmrequest.req_type = VMPTYPE_CHECK;
				caller->p_vmrequest.target = p[i]->p_endpoint;
				caller->p_vmrequest.params.check.start = v[i]->offset;
				caller->p_vmrequest.params.check.length = bytes;
				caller->p_vmrequest.params.check.writeflag = i;
				caller->p_vmrequest.nextrequestor = vmrequest;
				if (caller->p_vmrequest.nextrequestor == NULL) {
					send_sig(VM_PROC_NR, SIGKMEM);
				}
				vmrequest = caller;
				return VMSUSPEND;
			}
			if (s[0] < s[1]) {
				n = s[0];
			} else {
				n = s[1];
			}
			memcpy((void *)l[1], (void *)l[0], n);
			for (i = 0; i < 2; ++i) {
				l[i] += n;
				s[i] -= n;
			}
			o += n;
		}
		return 0;
	}
}

int data_copy(endpoint_t from, vir_bytes from_addr, endpoint_t to, vir_bytes to_addr, size_t bytes) {
	struct vir_addr f;
	struct vir_addr t;
	f.proc_nr_e = from;
	f.segment = D;
	f.offset = from_addr;
	t.proc_nr_e = to;
	t.segment = D;
	t.offset = to_addr;
	return virtual_copy_f(NULL, &f, &t, bytes, 0);
}

int data_copy_vmcheck(struct proc * caller, const endpoint_t from_proc, const vir_bytes from_addr, const endpoint_t to_proc, const vir_bytes to_addr, size_t bytes) {
	struct vir_addr f;
	struct vir_addr t;
	f.proc_nr_e = from_proc;
	f.segment = D;
	f.offset = from_addr;
	t.proc_nr_e = to_proc;
	t.segment = D;
	t.offset = to_addr;
	return virtual_copy_f(caller, &f, &t, bytes, 1);
}

void alloc_segments(struct proc *rp) {
	/* ToDo: MMU */
}

#if 0
void vm_stop(void) {
y_device_cons_put_string("{vm_stop}\n");
while (1) {
}
}
#else /* patch */
#endif

phys_bytes umap_local(register struct proc *rp, int seg, vir_bytes vir_addr, vir_bytes bytes) {
	if (seg == T) {
		if (vir_addr / CLICK_SIZE >= rp->p_memmap[T].mem_vir && (vir_addr + bytes - 1) / CLICK_SIZE < rp->p_memmap[T].mem_vir + rp->p_memmap[T].mem_len) {
			return vir_addr + (rp->p_memmap[T].mem_phys - rp->p_memmap[T].mem_vir) * CLICK_SIZE;
		}
	} else {
		if (vir_addr / CLICK_SIZE >= rp->p_memmap[T].mem_vir && (vir_addr + bytes - 1) / CLICK_SIZE < rp->p_memmap[T].mem_vir + rp->p_memmap[T].mem_len) { /* ToDo: .rodata */
			return vir_addr + (rp->p_memmap[T].mem_phys - rp->p_memmap[T].mem_vir) * CLICK_SIZE;
		}
		if (vir_addr / CLICK_SIZE >= rp->p_memmap[D].mem_vir && (vir_addr + bytes - 1) / CLICK_SIZE < rp->p_memmap[S].mem_vir + rp->p_memmap[S].mem_len) {
			return vir_addr + (rp->p_memmap[D].mem_phys - rp->p_memmap[D].mem_vir) * CLICK_SIZE;
		}
	}
	return 0;
}

#if 0
phys_bytes umap_virtual(struct proc* rp, int seg, vir_bytes vir_addr, vir_bytes bytes) {
y_device_cons_put_string("{umap_virtual}\n");
while (1) {
}
}
#else /* patch */
#endif

phys_bytes seg2phys(const u16_t seg) {
y_device_cons_put_string("{seg2phys}\n");
while (1) {
}
}

int vm_phys_memset(phys_bytes source, u8_t pattern, phys_bytes count) {
	/* ToDo: cache */
	memset((void *)source, pattern, count);
	return 0;
}

int intr_init(const int mine, const int auto_eoi) {
	return 0;
}

void halt_cpu(void) {
	yk0a_interrupt_t ignition;
	do {
		yk0a_flag_t f;
		yk0a_listen(1, glue_mask, &f);
		ignition = f & glue_mask;
	} while (ignition == 0);
	glue_interrupt(ignition);
}

void arch_init(void) {
}

void cpu_identify(void) {
}

void fpu_init(void) {
	/* ToDo: FPU */
}

#if 0
int is_fpu(void) {
y_device_cons_put_string("{is_fpu}\n");
while (1) {
}
}
#else /* patch */
#endif

#if 0
void ser_putc(char c) {
y_device_cons_put_string("{ser_putc}\n");
while (1) {
}
}
#else /* patch */
#endif

__dead void arch_shutdown(int how) {
y_device_cons_put_string("{arch_shutdown}\n");
while (1) {
}
}

#if 0
__dead void arch_monitor(void) {
y_device_cons_put_string("{arch_monitor}\n");
while (1) {
}
}
#else /* patch */
#endif

void restore_user_context(struct proc * p) {
}

void read_tsc(u32_t *high, u32_t *low) {
	/* ToDo: TSC */
	*high = 0;
	*low = 0;
}

#if 0
int arch_init_profile_clock(u32_t freq) {
y_device_cons_put_string("{arch_init_profile_clock}\n");
while (1) {
}
}
#else /* patch */
#endif

#if 0
void arch_stop_profile_clock(void) {
y_device_cons_put_string("{arch_stop_profile_clock}\n");
while (1) {
}
}
#else /* patch */
#endif

#if 0
void arch_ack_profile_clock(void) {
y_device_cons_put_string("{arch_ack_profile_clock}\n");
while (1) {
}
}
#else /* patch */
#endif

#if 0
void do_ser_debug(void) {
y_device_cons_put_string("{do_ser_debug}\n");
while (1) {
}
}
#else /* patch */
#endif

int arch_get_params(char *parm, int max) {
	strncpy(parm, glue_boot, max - 1);
	parm[max - 1] = '\0';
	return 0;
}

int arch_set_params(char *parm, int max) {
y_device_cons_put_string("{arch_set_params}\n");
while (1) {
}
}

void arch_pre_exec(struct proc *pr, const u32_t ip, const u32_t sp) {
	pr->p_reg.pc = ip;
	pr->p_reg.sp = sp;
}

int arch_umap(const struct proc *pr, vir_bytes offset, vir_bytes count, int seg, phys_bytes *addr) {
y_device_cons_put_string("{arch_umap}\n");
while (1) {
}
}

int arch_do_vmctl(message *m_ptr, struct proc *p) {
	/* ToDo: MMU */
	switch (m_ptr->SVMCTL_PARAM) {
	case VMCTL_INCSP:
		p->p_reg.sp += m_ptr->SVMCTL_VALUE;
		return 0;
		break;
	case VMCTL_FLUSHTLB:
		glue_vmctl = 1;
		return 0;
		break;
	case VMCTL_SETADDRSPACE:
		p->p_seg.p_cr3 = m_ptr->SVMCTL_PTROOT;
		p->p_misc_flags |= MF_FULLVM;
		RTS_UNSET(p, RTS_VMINHIBIT);
		return 0;
		break;
	default:
y_device_cons_put_string("{arch_do_vmctl}\n");
while (1) {
}
		break;
	}
}

int vm_contiguous(const struct proc *targetproc, vir_bytes vir_buf, size_t count) {
y_device_cons_put_string("{vm_contiguous}\n");
while (1) {
}
}

void proc_stacktrace(struct proc *proc) {
	/* ToDo: frame pointer */
}

int vm_lookup(const struct proc *proc, vir_bytes virtual, phys_bytes *result, u32_t *ptent) {
	if (proc->p_seg.p_cr3 == 0) {
		*result = virtual;
		return 0;
	} else {
y_device_cons_put_string("{vm_lookup}\n");
while (1) {
}
	}
}

void delivermsg(struct proc *target) {
	reg_t r;
	if (copy_msg_to_user(target, &target->p_delivermsg, target->p_delivermsg_vir) != 0) {
		r = EFAULT;
	} else {
		target->p_delivermsg.m_source = NONE;
		target->p_misc_flags &= ~MF_DELIVERMSG;
		r = 0;
	}
	if ((target->p_misc_flags & MF_CONTEXT_SET) == 0) {
		target->p_reg.retreg = r;
	}
}

void arch_do_syscall(struct proc *proc) {
y_device_cons_put_string("{arch_do_syscall}\n");
while (1) {
}
}

int arch_phys_map(int index, phys_bytes *addr, phys_bytes *len, int *flags) {
	return EINVAL;
}

int arch_phys_map_reply(int index, vir_bytes addr) {
y_device_cons_put_string("{arch_phys_map_reply}\n");
while (1) {
}
}

int arch_enable_paging(struct proc * caller, const message * m_ptr) {
	/* ToDo: MMU */
	int r;
	struct vm_ep_data v;
	vm_running = 1;
	r = data_copy(caller->p_endpoint, m_ptr->SVMCTL_VALUE, KERNEL, &v, sizeof(v));
	if (r != 0) {
		return r;
	}
	if (newmap(caller, caller, v.mem_map) != 0) {
		panic("arch_enable_paging");
	}
	return 0;
}

int copy_msg_from_user(struct proc * p, message * user_mbuf, message * dst) {
	/* ToDo: cache */
	phys_bytes l;
	l = umap_local(p, D, user_mbuf, sizeof(message));
	if (l == 0) {
		return -1;
	}
	if (p->p_seg.p_cr3 == 0) {
		memcpy(dst, l, sizeof(message));
		return 0;
	} else {
		phys_bytes o;
		o = 0;
		while (o < sizeof(message)) {
			y_uint32_t d;
			phys_bytes s;
			y_uint32_t t;
			d = *(y_uint32_t *)(p->p_seg.p_cr3 + I386_VM_PDE((y_uint32_t)user_mbuf + o) * 4);
			if ((d & I386_VM_PRESENT) == 0) {
				return -1;
			}
			t = *(y_uint32_t *)(I386_VM_PFA(d) + I386_VM_PTE((y_uint32_t)user_mbuf + o) * 4);
			if ((t & I386_VM_PRESENT) == 0) {
				return -1;
			}
			l = I386_VM_PFA(t) + ((y_uint32_t)user_mbuf + o) % I386_PAGE_SIZE;
			s = I386_PAGE_SIZE - ((y_uint32_t)user_mbuf + o) % I386_PAGE_SIZE;
			if (s > sizeof(message) - o) {
				s = sizeof(message) - o;
			}
			memcpy(dst + o, l, s);
			o += s;
		}
		return 0;
	}
}

int copy_msg_to_user(struct proc * p, message * src, message * user_mbuf) {
	/* ToDo: cache */
	phys_bytes l;
	l = umap_local(p, D, user_mbuf, sizeof(message));
	if (l == 0) {
		return -1;
	}
	if (p->p_seg.p_cr3 == 0) {
		memcpy(l, src, sizeof(message));
		return 0;
	} else {
		phys_bytes o;
		o = 0;
		while (o < sizeof(message)) {
			y_uint32_t d;
			phys_bytes s;
			y_uint32_t t;
			d = *(y_uint32_t *)(p->p_seg.p_cr3 + I386_VM_PDE((y_uint32_t)user_mbuf + o) * 4);
			if ((d & I386_VM_PRESENT) == 0) {
				return -1;
			}
			t = *(y_uint32_t *)(I386_VM_PFA(d) + I386_VM_PTE((y_uint32_t)user_mbuf + o) * 4);
			if ((t & I386_VM_PRESENT) == 0) {
				return -1;
			}
			if ((t & I386_VM_WRITE) == 0) {
				return -1;
			}
			l = I386_VM_PFA(t) + ((y_uint32_t)user_mbuf + o) % I386_PAGE_SIZE;
			s = I386_PAGE_SIZE - ((y_uint32_t)user_mbuf + o) % I386_PAGE_SIZE;
			if (s > sizeof(message) - o) {
				s = sizeof(message) - o;
			}
			memcpy(l, src + o, s);
			o += s;
		}
		return 0;
	}
}

#if 0
void switch_address_space(struct proc * p) {
y_device_cons_put_string("{switch_address_space}\n");
while (1) {
}
}
#else /* patch */
#endif

void release_address_space(struct proc *pr) {
	/* ToDo: MMU */
}

void enable_fpu_exception(void) {
	/* ToDo: FPU */
}

void disable_fpu_exception(void) {
y_device_cons_put_string("{disable_fpu_exception}\n");
while (1) {
}
}

#if 0
void release_fpu(struct proc * p) {
y_device_cons_put_string("{release_fpu}\n");
while (1) {
}
}
#else /* patch */
#endif

#if 0
void arch_pause(void) {
y_device_cons_put_string("{arch_pause}\n");
while (1) {
}
}
#else /* patch */
#endif

short cpu_load(void) {
y_device_cons_put_string("{cpu_load}\n");
while (1) {
}
}

#if 0
void busy_delay_ms(int ms) {
y_device_cons_put_string("{busy_delay_ms}\n");
while (1) {
}
}
#else /* patch */
#endif

#include "system.h"

int do_int86(struct proc * caller, message *m_ptr) {
y_device_cons_put_string("{do_int86}\n");
while (1) {
}
}

int do_sdevio(struct proc * caller, message *m_ptr) {
y_device_cons_put_string("{do_sdevio}\n");
while (1) {
}
}

int do_iopenable(struct proc * caller, message *m_ptr) {
y_device_cons_put_string("{do_iopenable}\n");
while (1) {
}
}

int do_readbios(struct proc * caller, message *m_ptr) {
y_device_cons_put_string("{do_readbios}\n");
while (1) {
}
}

#include <minix/portio.h>

unsigned inb(u16_t _port) {
y_device_cons_put_string("{inb}\n");
while (1) {
}
}

unsigned inw(u16_t _port) {
y_device_cons_put_string("{inw}\n");
while (1) {
}
}

unsigned inl(u16_t _port) {
y_device_cons_put_string("{inl}\n");
while (1) {
}
}

void outb(u16_t _port, u8_t _value) {
y_device_cons_put_string("{outb}\n");
while (1) {
}
}

void outw(u16_t _port, u16_t _value) {
y_device_cons_put_string("{outw}\n");
while (1) {
}
}

void outl(u16_t _port, u32_t _value) {
y_device_cons_put_string("{outl}\n");
while (1) {
}
}
